import os
import sys
from pathlib import Path


def load_path():
    """
    向上查找 [config.py or pyproject.toml] 所在目录，并将其添加到 sys.path 中。
    """
    current_dir = Path(sys.modules.get("__main__").__file__).resolve().parent
    root_dir = None
    for parent in [current_dir, *current_dir.parents]:
        if (parent / "config.py").exists() or (parent / "pyproject.toml").exists():
            root_dir = parent
            break
    if root_dir is not None:
        root_path = str(root_dir)
        if sys.path[0] != root_path:
            sys.path.insert(0, root_path)
            print(f"[load_path] 添加路径到 sys.path: {root_path}")
    else:
        print("[load_path] 未找到 config.py 或 pyproject.toml，未修改 sys.path。")



"""
@author: noybzy
@time: 2024/4/14 下午15:02
@file: database.py
@describe: http操作
"""
import abc
import time
from typing import Optional, TYPE_CHECKING

import requests
from loguru import logger

from .lazy import LazyLoader

if TYPE_CHECKING:
    import httpx
    import curl_cffi
else:
    httpx = LazyLoader("httpx")
    curl_cffi = LazyLoader("curl_cffi")


class RetryRequest(Exception):
    def __init__(self, flush=True):
        self.flush = flush


class BaseRequest(abc.ABC):

    def __init__(self, proxy_url: str):
        self.proxy_url = proxy_url
        self.proxy = None
        self.headers = None
        self.cookies = None
        self.timeout = 6
        self.verify = False
        self.req = None
        self.max_retry = 25
        self.flush_proxy()
        self.data = {} # 扩展字段

    def get_proxy(self) -> Optional[dict]:
        if self.proxy_url:
            num = 0
            while True:
                num += 1
                try:
                    proxy = requests.get(self.proxy_url, timeout=6).json()
                    break
                except Exception as e:
                    logger.warning(f'获取代理出错_{num}:{e.__class__}')
                    time.sleep(5)
            return proxy
        return None

    def flush_proxy(self):
        logger.debug(f'刷新代理: flush')
        self.proxy = self.get_proxy()

    @abc.abstractmethod
    def get(self, url: str, params=None, headers=None, cookies=None, allow_redirects=True, show_debug=True,
            timeout=None, auth=None, stream=None, cert=None, verify=None
            ):
        pass

    @abc.abstractmethod
    def post(self, url, params=None, data=None, json=None, allow_redirects=True, headers=None,
             cookies=None, show_debug=False,
             timeout=None, auth=None, stream=None, cert=None, verify=None
             ):
        pass

    def request_interceptor(self, url: str, params: str, data, json_data: Optional[dict], method: str):
        """请求拦截器"""
        pass

    def response_interceptor(self, response):
        """响应拦截器"""
        pass

    def exception_interceptor(self, exc):
        pass

    def put(self, *args):
        pass

    def delete(self, *args):
        pass


class Request(BaseRequest):

    def __init__(self, proxy_url: str, headers=None, cookies=None, timeout=6):
        super().__init__(proxy_url=proxy_url)
        self.req = requests
        self.headers = headers
        self.cookies = cookies
        self.timeout = timeout

    def exception_interceptor(self, e):
        try:
            raise e
        except requests.exceptions.ProxyError as e:
            logger.debug(f"代理错误: {e.__class__.__name__}")
            raise RetryRequest()
        except requests.exceptions.Timeout as e:
            logger.debug(f"超时: {e.__class__.__name__}")
            raise RetryRequest()
        except Exception as e:
            raise e

    def get(self, url: str, params=None, headers=None, cookies=None, allow_redirects=True, show_debug=True,
            timeout=None, auth=None, stream=None, cert=None, verify=None) -> Optional[requests.Response]:
        """
        发送get请求
        :param cookies:
        :param headers:
        :param allow_redirects:
        :param url:
        :param params:
        :param show_debug:
        :return:
        """
        num = 0
        # 请求拦截器
        while True:
            num = num + 1
            self.request_interceptor(url, params, None, None, 'get')

            final_cookies = cookies if cookies is not None else self.cookies
            final_headers = headers if headers is not None else self.headers
            final_timeout = timeout if timeout is not None else self.timeout
            final_verify = verify if verify is not None else self.verify

            if show_debug:
                logger.debug(f"发送请求GET_{num}：{url}")
            try:
                response = self.req.get(
                    url=url, proxies=self.proxy, cookies=final_cookies, headers=final_headers,
                    timeout=final_timeout, verify=final_verify, params=params, auth=auth,
                    stream=stream, cert=cert,
                    allow_redirects=allow_redirects
                )
                self.response_interceptor(response)
                return response
            except RetryRequest as e:
                if num > self.max_retry:
                    raise e
                if e.flush:
                    self.flush_proxy()
            except Exception as e:
                if num > self.max_retry:
                    raise e
                try:
                    self.exception_interceptor(e)
                except RetryRequest as e:
                    if e.flush:
                        self.flush_proxy()
                    continue
                except Exception as e:
                    raise e

    def post(self, url, params=None, data=None, json=None, allow_redirects=True, headers=None, cookies=None,
             show_debug=False, timeout=None, auth=None, stream=None, cert=None, verify=None) -> Optional[
        requests.Response]:
        """
        发送post请求
        :param allow_redirects:
        :param url:
        :param params:
        :param data:
        :param json_data:
        :return:
        """
        num = 0
        while True:
            num = num + 1
            self.request_interceptor(url, params, None, None, 'get')

            final_cookies = cookies if cookies is not None else self.cookies
            final_headers = headers if headers is not None else self.headers
            final_timeout = timeout if timeout is not None else self.timeout
            final_verify = verify if verify is not None else self.verify

            if show_debug:
                logger.debug(f"发送请求POST_{num}：{url}")
                if json:
                    logger.debug(f'json_data:{json}')
                if data:
                    logger.debug(f'data:{data}')

            try:
                response = self.req.post(url=url, proxies=self.proxy, cookies=final_cookies, headers=final_headers,
                                         timeout=final_timeout, verify=final_verify, params=params, data=data,
                                         json=json, allow_redirects=allow_redirects, stream=stream, cert=cert,
                                         auth=auth)
                self.response_interceptor(response)
                return response
            except RetryRequest as e:
                if num > self.max_retry:
                    raise e
                if e.flush:
                    self.flush_proxy()
            except Exception as e:
                if num > self.max_retry:
                    raise e
                try:
                    self.exception_interceptor(e)
                except RetryRequest as e:
                    if e.flush:
                        self.flush_proxy()
                    continue
                except Exception as e:
                    raise e


class RequestHttpx(BaseRequest):

    def __init__(self, proxy_url: str, headers=None, cookies=None, timeout=6):
        super().__init__(proxy_url=proxy_url)
        self.req = httpx
        self.headers = headers
        self.cookies = cookies
        self.timeout = timeout

    def exception_interceptor(self, e):
        try:
            raise e
        except httpx.ConnectError as e:
            logger.debug(f"连接错误: {e.__class__.__name__}")
            raise RetryRequest()
        except httpx.ProxyError as e:
            logger.debug(f"代理错误: {e.__class__.__name__}")
            raise RetryRequest()
        except httpx.TimeoutException as e:
            logger.debug(f"超时: {e.__class__.__name__}")
            raise RetryRequest()
        except Exception as e:
            raise e

    def get(self, url: str, params=None, headers=None, cookies=None, allow_redirects=True, show_debug=True,
            timeout=None, auth=None, stream=None, cert=None, verify=None) -> 'Optional[httpx.Response]':
        """
        发送get请求
        :param verify:
        :param cert:
        :param stream:
        :param auth:
        :param timeout:
        :param cookies:
        :param headers:
        :param allow_redirects:
        :param url:
        :param params:
        :param show_debug:
        :return:
        """
        num = 0
        # 请求拦截器
        while True:
            num = num + 1
            self.request_interceptor(url, params, None, None, 'get')

            final_cookies = cookies if cookies is not None else self.cookies
            final_headers = headers if headers is not None else self.headers
            final_timeout = timeout if timeout is not None else self.timeout
            final_verify = verify if verify is not None else self.verify
            final_proxy = self.proxy.get('https' if url.startswith('https') else 'http') if self.proxy else None
            if show_debug:
                logger.debug(f"发送请求GET_{num}：{url}")
            try:
                response = self.req.get(
                    url=url, proxy=final_proxy, cookies=final_cookies, headers=final_headers,
                    timeout=final_timeout, verify=final_verify, params=params, auth=auth,
                    follow_redirects=allow_redirects
                )
                self.response_interceptor(response)
                return response
            except RetryRequest as e:
                if num > self.max_retry:
                    raise e
                if e.flush:
                    self.flush_proxy()
            except Exception as e:
                if num > self.max_retry:
                    raise e
                try:
                    self.exception_interceptor(e)
                except RetryRequest as e:
                    if e.flush:
                        self.flush_proxy()
                    continue
                except Exception as e:
                    raise e

    def post(self, url, params=None, data=None, json=None, allow_redirects=True, headers=None,
             cookies=None, show_debug=False,
             timeout=None, auth=None, stream=None, cert=None, verify=None
             ) -> 'Optional[httpx.Response]':
        """
       发送post请求
       :param allow_redirects:
       :param url:
       :param params:
       :param data:
       :param json_data:
       :return:
       """
        num = 0
        while True:
            num = num + 1
            self.request_interceptor(url, params, None, None, 'get')

            final_cookies = cookies if cookies is not None else self.cookies
            final_headers = headers if headers is not None else self.headers
            final_timeout = timeout if timeout is not None else self.timeout
            final_verify = verify if verify is not None else self.verify
            final_proxy = self.proxy.get('https' if url.startswith('https') else 'http') if self.proxy else None
            if show_debug:
                logger.debug(f"发送请求POST_{num}：{url}")
                if json:
                    logger.debug(f'json_data:{json}')
                if data:
                    logger.debug(f'data:{data}')

            try:
                response = self.req.post(url=url, proxy=final_proxy, cookies=final_cookies, headers=final_headers,
                                         timeout=final_timeout, verify=final_verify, params=params, data=data,
                                         json=json, follow_redirects=allow_redirects,
                                         auth=auth)
                self.response_interceptor(response)
                return response
            except RetryRequest as e:
                if num > self.max_retry:
                    raise e
                if e.flush:
                    self.flush_proxy()
            except Exception as e:
                if num > self.max_retry:
                    raise e
                try:
                    self.exception_interceptor(e)
                except RetryRequest as e:
                    if e.flush:
                        self.flush_proxy()
                    continue
                except Exception as e:
                    raise e


class RequestCurlCffi(Request):

    def __init__(self, proxy_url: str, headers=None, cookies=None, timeout: int = 6):
        super().__init__(proxy_url=proxy_url, headers=headers, cookies=cookies, timeout=timeout)
        self.req = curl_cffi.requests

    def exception_interceptor(self, e):
        try:
            raise e
        except curl_cffi.exceptions.ConnectionError as e:
            logger.debug(f"连接错误: {e.__class__.__name__}")
            raise RetryRequest()
        except curl_cffi.exceptions.ProxyError as e:
            logger.debug(f"代理错误: {e.__class__.__name__}")
            raise RetryRequest()
        except curl_cffi.exceptions.Timeout as e:
            logger.debug(f"超时: {e.__class__.__name__}")
            raise RetryRequest()
        except Exception as e:
            raise e

    def get(self, url: str, params=None, headers=None, cookies=None, allow_redirects=True, show_debug=True,
            timeout=None, auth=None, stream=None, cert=None, verify=None) -> 'Optional[curl_cffi.requests.Response]':
        return super().get(url=url, params=params, headers=headers, cookies=cookies, allow_redirects=allow_redirects, show_debug=show_debug, timeout=timeout, auth=auth, stream=stream, cert=cert, verify=verify)

    def post(self, url, params=None, data=None, json=None, allow_redirects=True, headers=None, cookies=None,
             show_debug=False, timeout=None, auth=None, stream=None, cert=None, verify=None) -> 'Optional[curl_cffi.requests.Response]':
        return super().post(url=url, params=params, data=data, json=json, allow_redirects=allow_redirects, headers=headers, cookies=cookies, show_debug=show_debug, timeout=timeout, auth=auth, stream=stream, cert=cert, verify=verify)



class MysqlConfig(object):
    def __init__(self, host: str, port: int, username: str, password: str, db: str, charset: str = 'utf8mb4'):
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.db = db
        self.charset = charset


class RedisConfig(object):
    def __init__(self, host: str, port: int, password: str, db: int):
        self.host = host
        self.port = port
        self.password = password
        self.db = db


class RabbitMQConfig(object):
    def __init__(self, host: str, port: int, username: str, password: str, virtual_host: str = '/'):
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.virtual_host = virtual_host


class OssConfig(object):
    def __init__(self, access_key: str, access_secret: str, endpoint: str, bucket_name: str, region: str = None):
        self.access_key = access_key
        self.access_secret = access_secret
        self.endpoint = endpoint
        self.bucket_name = bucket_name
        self.region = region



"""
@author: noybzy
@time: 2024/9/19 上午1:02
@file: database.py
@describe: 数据库操作 mysql，redis
@updated: 2025/2/14
"""
import json
import time
from typing import Union, TYPE_CHECKING

import pymysql
import pymysql.cursors
from dbutils.pooled_db import PooledDB
from loguru import logger

from .config import MysqlConfig, RedisConfig
from .lazy import LazyLoader

if TYPE_CHECKING:
    import redis
else:
    redis = LazyLoader("redis")


class ConnManager:
    def __init__(self, sql_pool, kv=False):
        self.kv = kv
        self.sql_pool = sql_pool
        self.conn: Union[pymysql.connections.Connection, None] = None
        self.cursor: Union[pymysql.cursors.Cursor, None] = None

    def __enter__(self) -> 'ConnManager':
        self.conn = self.sql_pool.connection()
        if self.kv:
            self.cursor = self.conn.cursor(cursor=pymysql.cursors.DictCursor)
        else:
            self.cursor = self.conn.cursor()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.cursor:
            self.cursor.close()
        if self.conn:
            self.conn.close()


class Mysql(object):
    def __init__(self, host: str = '127.0.0.1', username: str = 'root', password: str = '123456', db: str = 'test',
                 drop_column: list = None, pool_num: int = 4, max_pool: int = None, port: int = 3306, monitor=None):
        # monitor 历史遗留问题，后续计划删掉
        if drop_column is None:
            drop_column = ["id", "updated", 'created', 'isonline', 'islocal']
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.db = db
        self.drop_column = drop_column
        self.pool_num = pool_num
        self.sql_pool = PooledDB(
            creator=pymysql,  # 使用链接数据库的模块
            maxconnections=max_pool,  # 连接池允许的最大连接数，0和None表示不限制连接数
            mincached=self.pool_num,  # 初始化时，连接池中至少创建的空闲的连接，0表示不创建
            # maxcached=5,  # 连接池中空闲的最多连接数，0和None表示不限制
            # maxshared=3,  # 连接池中最多共享的连接数量，0和None表示全部共享
            blocking=True,  # 连接池中如果没有可用连接后，阻塞等待，而不是报错
            # maxusage=None,  # 一个连接最多被重复使用的次数，None表示无限制
            # setsession=[],  # 开始会话前执行的命令列表
            # ping=0,  # ping MySQL服务端，检查服务是否可用
            host=self.host,
            port=self.port,
            user=self.username,
            password=self.password,
            database=self.db,
            charset='utf8mb4'
        )
        self.sql_pool.connection()
        self.table_name_list = self.get_db_name()  # 获取所有表名
        self.column_list = {}

    @staticmethod
    def simple(config: MysqlConfig, pool_num=4, db: str = None, monitor=None, max_pool: int = None, ) -> 'Mysql':
        """
        创建Mysql
        :param max_pool: 最大连接数
        :param config: 连接配置
        :param pool_num: 线程池数量
        :param db: 数据库
        :param monitor: 监控
        :return:
        """

        if db is None:
            db = config.db
        mysql = Mysql(
            host=config.host,
            port=config.port,
            username=config.username,
            password=config.password,
            db=db,
            pool_num=pool_num,
            max_pool=max_pool,
        )
        return mysql

    def get_db_name(self, db: str = None) -> list:
        """
        获取指定库的所有表
        :return:
        """
        if db is None:
            db = self.db

        sql = "select table_name from information_schema.tables where table_schema='{}'".format(db)
        with ConnManager(self.sql_pool) as cm:
            cm.cursor.execute(sql)
            db_list = cm.cursor.fetchall()
            db_list = [i[0] for i in db_list]
            return db_list

    def get_table_column_list(self, table_name: str) -> dict:
        """
        获取指定表的所有字段
        :param table_name: 表名
        :return:
        """
        column_list = self.column_list.get(table_name)
        if column_list:
            return column_list
        else:
            sql = 'select column_name,data_type from information_schema.columns where table_name=%s and table_schema=%s'
            with ConnManager(self.sql_pool) as cm:
                cm.cursor.execute(sql, (table_name, self.db))
                column_list = cm.cursor.fetchall()
                insert_columns = {}
                for i in column_list:
                    if i[0] in self.drop_column:
                        continue
                    insert_columns[i[0]] = i[1]
                self.column_list[table_name] = insert_columns
                return insert_columns

    def execute_sql(self, sql, params=None, kv=True) -> Union[list, int]:
        """
        执行sql语句
        :param sql: sql代码
        :param params: 参数
        :param kv: 查询结果是否转成字典
        :return: 返回查询结果 | 受影响行数
        """
        try:
            with ConnManager(self.sql_pool, kv=kv) as cm:
                cm.cursor.execute(sql, params)
                if sql.strip().lower().startswith("select"):
                    result = cm.cursor.fetchall()  # 查询语句返回所有结果
                    return result
                else:
                    # 非查询语句，返回受影响的行数
                    cm.conn.commit()
                    return cm.cursor.rowcount
        except Exception as e:
            logger.error(f'执行sql失败：{e.__class__.__name__}|{e.args}>>{sql}, {params}')
            raise e

    @staticmethod
    def wrapper_value(key, value, insert_columns):
        if value is None:
            if insert_columns.get(key) in ('datetime', 'date', 'time', 'year'):
                return None
            else:
                return ''
        else:
            if isinstance(value, bool):
                return value
            else:
                return str(value)

    def insert_data(self, item: dict, table_name: str, replace: bool = False) -> int:
        """
        插入单条数据(默认表里的字段不允许为空[除 'datetime', 'date', 'time', 'year' 字段],['bigint','int', 'decimal' 字段为None时取得默认值] )
        :param item: 数据
        :param table_name: 表名
        :param replace: 是否使用replace，默认insert
        :return: 返回成功的行数
        """
        if item:
            insert_columns = self.get_table_column_list(table_name)
            # 过滤字段，并移除 'int' 或 'decimal' 类型为 None 的字段 (采用数据库默认值)
            filtered_item = {
                k: v for k, v in item.items() if
                k in insert_columns.keys() and not (v is None and insert_columns[k] in ('bigint', 'int', 'decimal','tinyint'))
            }
            if not filtered_item:
                logger.warning(f"没有找到可插入的字段: {item}")
                return 0

            insert_columns_new = ['`%s`' % i for i in filtered_item.keys()]
            # 选择 SQL 操作：INSERT 或 REPLACE
            operation = "REPLACE" if replace else "INSERT"
            columns = ', '.join(insert_columns_new)
            placeholders = ', '.join(['%s'] * len(filtered_item))
            sql = f"{operation} INTO {table_name} ({columns}) VALUES ({placeholders})"

            values_list = []
            for key in filtered_item.keys():
                value = filtered_item.get(key)
                values_list.append(self.wrapper_value(key, value, insert_columns))
            values = tuple(values_list)

            try:
                with ConnManager(self.sql_pool) as cm:
                    cm.cursor.execute(sql, values)
                    cm.conn.commit()  # 提交事务
                    logger.success(f'{operation} {table_name} rows:{cm.cursor.rowcount} successfully')
                    rowcount = cm.cursor.rowcount
                    return rowcount
            except pymysql.IntegrityError as e:
                error_code, error_message = e.args  # 获取错误码和错误信息
                if error_code == 1062:
                    # 提取唯一键冲突的详细信息
                    duplicate_key = error_message.split("for key")[1].strip().replace('\'', '').replace('\"',
                                                                                                        '') if "for key" in error_message else "Unknown key"
                    duplicate_value = item.get(duplicate_key.split('.')[-1])
                    logger.warning(f"IntegrityError: {duplicate_key}: {duplicate_value}")
                else:
                    logger.warning(f"IntegrityError: {e.args}")
                return 0
            except pymysql.ProgrammingError as e:
                logger.error(f'sql 语法错误 {e.__class__.__name__},sql:{sql}')
                raise e
            except Exception as e:
                logger.error(f'insert error {e.__class__.__name__}')
                raise e

        else:
            logger.warning('插入为空')

    def insert_data_many(self, item_list: list[dict], table_name: str, batch_size: int = 500, replace: bool = False,
                         ignore: bool = False) -> int:
        """
        批量插入数据(默认表里的字段不允许为空[除 'datetime', 'date', 'time', 'year' 字段],注意['bigint','int', 'decimal' 字段为None时请手动设置默认值] )
        :param item_list: 数据列表
        :param table_name: 表名
        :param batch_size: 批量插入大小
        :param replace: 是否使用replace，默认insert
        :param ignore: 是否忽略掉插入错误
        :return:
        """
        # 批量插入不过滤插入参数

        if not item_list:
            logger.warning("插入的数据列表为空")
            return 0

        # 获取表的列名
        insert_columns = self.get_table_column_list(table_name)
        if not insert_columns:
            logger.warning(f"表 {table_name} 的列名列表为空")
            return 0

        insert_columns_new = ['`%s`' % i for i in insert_columns.keys()]
        # 生成批量插入的 SQL 和数据
        operation = "REPLACE" if replace else "INSERT"
        if ignore:
            operation = operation + ' ' + 'ignore '
        columns = ', '.join(insert_columns_new)
        placeholders = ', '.join(['%s'] * len(insert_columns))
        sql = f"{operation} INTO {table_name} ({columns}) VALUES ({placeholders})"

        rows = []
        for item in item_list:
            values_list = []
            for key in insert_columns.keys():
                value = item.get(key)
                values_list.append(self.wrapper_value(key, value, insert_columns))
            rows.append(tuple(values_list))
        total_inserted = 0
        try:
            with ConnManager(self.sql_pool) as cm:
                # 批量插入，每次插入 batch_size 条记录
                for i in range(0, len(rows), batch_size):
                    batch = rows[i:i + batch_size]
                    try:
                        cm.cursor.executemany(sql, batch)  # 批量插入
                        cm.conn.commit()  # 提交事务
                        inserted_rows = cm.cursor.rowcount
                        total_inserted += inserted_rows
                        logger.success(f"insert {inserted_rows}/{len(batch)} rows into {table_name} successfully.")
                    except pymysql.IntegrityError as e:
                        # 捕获批量插入时的唯一键冲突错误
                        logger.warning(f"IntegrityError during batch insert,尝试单条插入")
                        cm.conn.rollback()
                        for item in item_list[i:i + batch_size]:
                            try:
                                # 逐条插入调用 insert_data 方法
                                self.insert_data(item, table_name, replace)
                            except Exception as individual_e:
                                logger.error(f"insert error during individual inserts: {individual_e}")
                                raise individual_e
        except pymysql.ProgrammingError as e:
            logger.error(f"SQL 语法错误: {e}, SQL: {sql}")
            raise e
        except Exception as e:
            logger.error(f"Insert error: {e.__class__.__name__} - {str(e)}")
            raise e
        return total_inserted


class Redis(object):
    def __init__(self, host: str, port: int, password: str, db: int, decode_responses: bool = False):
        self.host = host
        self.port = port
        self.password = password
        self.db = db
        self.r = redis.StrictRedis(host=host, port=port, password=password, db=db, decode_responses=decode_responses)

    @staticmethod
    def simple(config: RedisConfig, db: int = None, decode_responses: bool = True) -> 'Redis':
        """
        使用配置创建Redis对象
        :param config: 配置
        :param db: 数据库
        :param decode_responses: 是否自动解析字节
        :return:
        """
        if not db:
            db = config.db
        return Redis(
            host=config.host,
            port=config.port,
            password=config.password,
            db=db,
            decode_responses=decode_responses
        )

    def push(self, queue_name: str, data: Union[dict, str]):
        if isinstance(data, dict):
            data_str = json.dumps(data, ensure_ascii=False)
            self.r.lpush(queue_name, data_str)
        else:
            self.r.lpush(queue_name, data)

    def pop(self, queue_name: str, decode_json: bool = False) -> Union[dict, str]:
        data_str = self.r.rpop(queue_name)
        if decode_json:
            return json.loads(data_str)
        return data_str

    def delay_push(self, queue_name: str, data: Union[dict, str], delay: int = 125):
        """
        延迟添加数据到队列(配合delay_pop用)
        :param queue_name: 队列名称
        :param data: 数据
        :param delay: 延迟执行时间（单位：秒）
        """
        if isinstance(data, dict):
            data_str = json.dumps(data, ensure_ascii=False)
        else:
            data_str = data
        execute_time = time.time() + delay  # 计算任务的执行时间戳
        self.r.zadd(queue_name, {data_str: execute_time})

    def delay_pop(self, queue_name: str, delay=125, wait: bool = False) -> Union[str, None]:
        """
        延迟获取队列数据
        :param queue_name: 队列名称
        :param delay: 延迟时间
        :param wait: 是否等待获取
        :return:
        """
        while True:
            current_time = time.time()
            task_info = self.r.zpopmin(queue_name, 1)
            if task_info:
                task, score = task_info[0]
                if score <= current_time:
                    new_time = current_time + delay
                    self.r.zadd(queue_name, {task: new_time})
                    return task
                else:
                    self.r.zadd(queue_name, {task: score})
                    if wait:
                        logger.debug(f'等待队列 {queue_name}')
                        time.sleep(5)
                        continue
                    else:
                        return None
            else:
                return None



import datetime
import os.path
import time
import zipfile
from enum import Enum
import random
from typing import TYPE_CHECKING, Optional
from urllib.parse import unquote

from loguru import logger
import requests
from .config import OssConfig
from .http import Request
from .lazy import LazyLoader
from .utils import get_md5
from requests.exceptions import ProxyError, Timeout

requests.packages.urllib3.disable_warnings()
if TYPE_CHECKING:
    import oss2
else:
    oss2 = LazyLoader("oss2")


class UploadFile(object):
    """
    文件上传
    """

    def __init__(self, prefix: str, access_key: str = "", access_secret: str = "", endpoint: str = "", bucket_name: str = "", retry_num: int = 5):
        auth = oss2.Auth(access_key, access_secret)
        self.prefix = prefix
        self.bucket = oss2.Bucket(auth, endpoint, bucket_name)
        self.oss_path = prefix + '/' + datetime.datetime.now().strftime("%Y%m/%d")
        self.retry_num = retry_num
        logger.debug(f'当前oss根路径:{self.oss_path}')

    @staticmethod
    def simple(config: OssConfig, prefix: str, bucket_name: str = 'bainiudata1') -> 'UploadFile':
        """
        通过配置文件创建对象
        :param config:
        :param prefix:
        :param bucket_name:
        :return:
        """
        return UploadFile(
            prefix=prefix,
            access_key=config.access_key,
            access_secret=config.access_secret,
            endpoint=config.endpoint,
            bucket_name=bucket_name  # 存储桶
        )

    def is_exist(self, osspath):
        return self.bucket.object_exists(osspath)

    def upload(self, filepath: str, filename: str = None) -> str:
        self.oss_path = self.prefix + '/' + datetime.datetime.now().strftime("%Y%m/%d")
        if not filepath or not os.path.exists(filepath):
            logger.warning(f"本地文件不存在: {filepath}")
            return ''
        if filename:
            oss_abs_path = self.oss_path + '/' + filename
        else:
            oss_abs_path = self.oss_path + '/' + filepath.split('/')[-1]

        exist = self.bucket.object_exists(oss_abs_path)
        if exist:
            logger.warning(f"oss 存在相同名的文件 {oss_abs_path}")
            return oss_abs_path
        else:
            num = 0
            while True:
                try:
                    result = self.bucket.put_object_from_file(oss_abs_path, filepath)
                    break
                except Exception as e:
                    num += 1
                    logger.error(f'{e.__class__}')
                    if num > self.retry_num:
                        logger.error(f"oss upload faild {oss_abs_path},error:{e}")
                        return ''

            if int(result.status) == 200:
                logger.success(f"{oss_abs_path} 上传成功")
            else:
                logger.error(f"oss upload faild {oss_abs_path},status:{result.status}")
            return oss_abs_path


class DownloadFile(object):
    """
    下载文件
    """

    class NameMode(Enum):
        URL = 1
        Content_Disposition = 2
        Custom = 3

    def __init__(self, url, method='get', data=None, json_data: dict = None, headers: dict = None, cookies: dict = None, timeout: int = 20, proxy=False, save_path='file/', retry_num=5, name_mode=NameMode.URL,proxy_url=None):
        self.url = url
        self.headers = headers
        self.cookies = cookies
        self.proxy = proxy
        self.timeout = timeout
        self.method = method
        self.data = data
        self.json_data = json_data

        self.save_path = save_path  # 文件保存位置
        self.response = None  # 响应体
        self.filename = None  # 文件名
        self.filepath = None  # 文件路径
        self.prefix = None  # 文件前缀
        self.suffix = None  # 文件后缀
        self.name_mode: DownloadFile.NameMode = name_mode  # 文件名模式
        self.retry_num = retry_num  # 下载重试次数

        self.proxy_url = proxy_url


    def get_proxy(self) -> Optional[dict]:
        if self.proxy_url:
            num = 0
            while True:
                num += 1
                try:
                    proxy = requests.get(self.proxy_url, timeout=6).json()
                    break
                except Exception as e:
                    logger.warning(f'获取代理出错_{num}:{e.__class__}')
                    time.sleep(5)
            return proxy
        return None
    # def get_filename(self):
    #     """
    #     获取文件名
    #     模式：
    #         1，通过url获取
    #         2. 通过content-type获取
    #         3. 指定文件后缀
    #     :return:
    #     """
    #
    #     if self.name_mode == DownloadFile.NameMode.URL:
    #         self.suffix = '.' + self.url.split('?')[0].split('.')[-1]
    #         self.prefix = get_md5(self.url)
    #     if self.name_mode == DownloadFile.NameMode.Content_Disposition:
    #         dis = self.response.headers.get('Content-Disposition')
    #         if dis:
    #             filename = re.findall('filename=(.*)', dis)
    #             if filename:
    #                 self.suffix = filename[0].strip('\"').split('.')[-1]
    #                 self.suffix = '.' + self.suffix
    #         if not self.suffix:
    #             self.suffix = self.response.headers.get('Content-Type').split('/')[-1].split(';')[0]
    #             self.suffix = '.' + self.suffix
    #         self.prefix = get_md5(self.url)
    #     if self.name_mode == DownloadFile.NameMode.Custom:
    #         if self.filename:
    #             self.filepath = os.path.join(self.save_path, self.filename)
    #             return
    #         else:
    #             self.prefix = get_md5(self.url)
    #             if self.suffix is None:
    #                 logger.warning(f"未指定文件后缀")
    #                 self.suffix = ''
    #     self.filename = f"{self.prefix}{self.suffix}"
    #     self.filepath = os.path.join(self.save_path, self.filename)
    def get_filename(self):
        """
        根据指定模式生成文件名并构建完整存储路径
        模式：
            1，通过url获取
            2. 通过content-type获取
            3. 指定文件后缀
        :return:
        """

        if self.name_mode == DownloadFile.NameMode.URL:
            try:
                path_part = self.url.split('?', 1)[0]
                filename_part = path_part.split('/')[-1]
                if '.' in filename_part:
                    self.suffix = '.' + filename_part.rsplit('.', 1)[-1]
                    if len(self.suffix) > 10:
                        self.suffix = ''
                else:
                    self.suffix = ''  # 无有效扩展名时留空
                self.prefix = get_md5(self.url)
            except Exception as e:
                logger.exception(e)
                self.suffix = ''
                self.prefix = get_md5(self.url)
        elif self.name_mode == DownloadFile.NameMode.Content_Disposition:
            # 从Content-Disposition头解析文件名
            content_disposition = self.response.headers.get('Content-Disposition', '')
            filename = self._parse_content_disposition(content_disposition)
            if filename:
                if '.' in filename:
                    self.suffix = '.' + filename.rsplit('.', 1)[-1]
                else:
                    # 无扩展名时从Content-Type获取类型
                    self.suffix = self._get_suffix_from_content_type()
            else:
                # 无法解析文件名时从Content-Type获取
                self.suffix = self._get_suffix_from_content_type()
            self.prefix = get_md5(self.url)
        elif self.name_mode == DownloadFile.NameMode.Custom:
            if self.filename:
                # 直接使用用户指定的文件名
                self.filepath = os.path.join(self.save_path, self.filename)
                return
            else:
                # 生成基础文件名
                self.prefix = get_md5(self.url)
                if self.suffix is None:
                    logger.warning("未指定文件后缀，将使用空后缀")
                    self.suffix = ''
        else:
            raise ValueError(f"未知的文件命名模式：{self.name_mode}")

        # 构建最终文件名和路径
        self.filename = f"{self.prefix}{self.suffix}"
        self.filepath = os.path.join(self.save_path, self.filename)

    def _parse_content_disposition(self, content_disposition):
        """解析Content-Disposition头获取文件名"""
        if not content_disposition:
            return None

        filename = None
        parts = [p.strip() for p in content_disposition.split(';')]

        for part in parts:
            if part.lower().startswith('filename='):
                filename = part[len('filename='):].strip()
                # 去除包裹的引号
                if filename.startswith(('"', "'")):
                    filename = filename[1:-1]
                filename = unquote(filename)
                break
            elif part.lower().startswith('filename*='):
                filename_part = part[len('filename*='):]
                try:
                    encoding, _, filename_encoded = filename_part.split("'", 2)
                    filename = unquote(filename_encoded)
                    break
                except ValueError:
                    continue

        return filename

    def _get_suffix_from_content_type(self):
        """从Content-Type头提取文件后缀"""
        # MIME_MAPPING = {
        #     # 常见文档类型
        #     'application/pdf': '.pdf',
        #     'application/msword': '.doc',
        #     'application/vnd.ms-excel': '.xls',
        #     'application/vnd.openxmlformats-officedocument.wordprocessingml.document': '.docx',
        #     'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet': '.xlsx',
        #     'application/vnd.ms-powerpoint': '.ppt',
        #
        #     # 图片类型
        #     'image/jpeg': '.jpg',
        #     'image/png': '.png',
        #     'image/gif': '.gif',
        #     'image/webp': '.webp',
        #     'image/svg+xml': '.svg',
        #
        #     # 压缩文件
        #     'application/zip': '.zip',
        #     'application/x-rar-compressed': '.rar',
        #     'application/x-tar': '.tar',
        #     'application/gzip': '.gz',
        #
        #     # 文本类型
        #     'text/plain': '.txt',
        #     'text/csv': '.csv',
        #     'text/html': '.html',
        #     'application/json': '.json',
        #     'application/xml': '.xml',
        #
        #     # 二进制流
        #     'application/octet-stream': ''
        # }
        content_type = self.response.headers.get('Content-Type', 'application/octet-stream').split(';', 1)[0]
        mime_type = content_type.split('/', 1)[-1] if '/' in content_type else 'octet-stream'
        return '.' + mime_type

    def download(self, mode='stream'):
        logger.debug(f"下载文件：{self.url}")
        num = 0
        while True:
            try:
                if self.method == 'get':
                    response = requests.get(url=self.url, headers=self.headers, cookies=self.cookies, timeout=self.timeout,
                                            verify=False, proxies=self.proxy)
                elif self.method == 'post':
                    response = requests.post(url=self.url, data=self.data, json=self.json_data, headers=self.headers,
                                             cookies=self.cookies, timeout=self.timeout, verify=False, proxies=self.proxy)
                self.response = response

                if response.status_code == 404:
                    logger.warning(f"下载文件出错code:{response.status_code},放弃下载")
                    break

                if response.status_code != 200:
                    num = num + 1
                    logger.warning(f"下载文件出错code:{response.status_code} :retry：{num}")
                    if num > self.retry_num:
                        logger.error(f"下载失败：{self.url},达到最大重试次数，放弃下载")
                        break
                    time.sleep(3)
                    continue
            except ProxyError as e:
                self.proxy = self.get_proxy()
                logger.debug(f'{e.__class__}')
                continue
            except Exception as e:
                num += 1
                logger.error(f"{e}")
                logger.warning(f"下载文件出错：{self.url},retry：{num}")
                if num > self.retry_num:
                    logger.error(f"下载失败：{self.url},达到最大重试次数，放弃下载")
                    break
                time.sleep(random.randint(3, 5))
                continue

            if not os.path.exists(self.save_path):
                os.makedirs(self.save_path)
            self.get_filename()
            with open(self.filepath, "wb") as f:
                f.write(response.content)
            logger.success(f"下载文件成功：{self.filepath}")
            break

    def remove(self):
        if not self.filepath:
            return
        if os.path.exists(self.filepath):
            os.remove(self.filepath)


def unzip(filepath, target='file', buffer_size=1024 * 64):
    """
    解压zip文件
    :param filepath:
    :param target:
    :param buffer_size:
    :return:
    """
    file_list = []
    with zipfile.ZipFile(filepath, 'r') as zip_file:
        for file_info in zip_file.infolist():
            filename = file_info.filename.encode('cp437').decode('gbk')
            file_path = os.path.join(target, filename)
            if file_info.is_dir():
                os.makedirs(file_path, exist_ok=True)
            else:
                os.makedirs(os.path.dirname(file_path), exist_ok=True)
                with zip_file.open(file_info) as source_file:
                    with open(file_path, "wb") as target_file:
                        while True:
                            buffer = source_file.read(buffer_size)
                            if not buffer:
                                break
                            target_file.write(buffer)
                file_list.append(file_path)
    return file_list


def get_file_content(filepath: str) -> str:
    with open(filepath, 'r', encoding='UTF-8') as file:
        result = file.read()
    return result



import atexit
import json
import os
import sys
import threading
import time
import uuid
import datetime
from pathlib import Path
import traceback
from typing import Optional

import requests
from loguru import logger
import socket


MONITOR_ENDPOINT = "http://127.0.0.1:14000/endpoint"
LOG_PATH = Path("C:\\logs") if sys.platform == "win32" else Path("/logs")
MAX_RETRY = 100
MONITOR_INTERVAL = 10

class SpiderInfo:
    def __init__(self, name=None, group_name=None, monitor=True,task_type=0, show_debug=False,monitor_endpoint='http://127.0.0.1:14000/endpoint'):
        self._monitor_endpoint = monitor_endpoint
        self.pid = str(os.getpid())
        self.uid = str(uuid.uuid4())
        self.parent_uid = None
        self.name = name
        self.group_name = group_name
        self.start_time = datetime.datetime.now()
        self.end_time = None
        self.run_time = None
        self.log_file = None
        self.file = None
        self.interpreter = sys.executable
        self.status = 0
        self.insert_count = 0
        self.progress = 0
        self.total_progress = 0
        self.exception = None
        self.exception_stack = None
        self.msg = ''
        self.data = ''
        self.task_type = task_type
        self.server = os.getenv("SERVER_NAME")
        self._show_debug = show_debug
        self._is_monitor = monitor
        self._today = datetime.date.today()
        self._lock = threading.Lock()
        self.lan_ip = '127.0.0.1'
        self.wan_ip = '0.0.0.0'
        self._monitor_thread: Optional[threading.Thread] = None
        self._stop_event = threading.Event()

        sys.excepthook = self.__global_exception_handler
        atexit.register(self.__before_exit)

        self.__initialize_log_file()

        if self._is_monitor:
            self.start_monitor()
            self.notice()

        self.__init_ip()


    def __initialize_log_file(self):
        self.__get_script_info()
        if self.log_file:
            logger.add(self.log_file, rotation="50 MB")
        self.status = 1

    def __get_script_info(self):
        main_module = sys.modules.get('__main__')
        filepath = main_module.__file__
        self.file = filepath
        self.log_file = self.__get_log_path(filepath)

    def __get_log_path(self, file_path: str) -> str:
        # 基础日志目录
        base_log_path = LOG_PATH
        original_path = Path(file_path).resolve()
        stem_name = original_path.stem
        timestamp = datetime.datetime.now().strftime("%m%d-%H%M%S.%f")[:-3].replace('.', '')
        log_filename = f"{stem_name}-{timestamp}.log"

        # 获取脚本路径相对于根目录的结构
        rel_path = original_path.parent.relative_to(original_path.anchor)
        full_path = base_log_path / rel_path / log_filename
        full_path.parent.mkdir(parents=True, exist_ok=True)
        return str(full_path)

    def __global_exception_handler(self, exc_type, exc_value, exc_traceback):
        self.debug('全局异常捕获')
        stack_trace = ''.join(traceback.format_exception(exc_type, exc_value, exc_traceback))
        logger.error(f"发生异常: {exc_type.__name__}: {exc_value}\n堆栈信息:\n{stack_trace}")
        self.exception_stack = stack_trace
        self.exception = f"{exc_type}:{exc_value}"
        self.status = -1
        sys.__excepthook__(exc_type, exc_value, exc_traceback)

    def __before_exit(self):
        self.debug('程序退出')
        self.end_time = datetime.datetime.now()
        self.run_time = self.end_time - self.start_time
        if self.status != -1:
            self.status = 2
        logger.warning(f'运行时长: {self.run_time} 秒')
        if self._is_monitor:
            self.notice()
        logger.warning(self)

    def __init_ip(self):
        try:
            s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            s.connect(("223.5.5.5", 80))
            self.lan_ip = s.getsockname()[0]
            s.close()
        except Exception:
            try:
                self.lan_ip = socket.gethostbyname(socket.gethostname())
            except:
                pass
        self.wan_ip = os.getenv("WAN_IP", '0.0.0.0')
        if self.wan_ip == '0.0.0.0':
            try:
                response = requests.get('https://ipinfo.io/json', timeout=5)
                self.wan_ip = response.json()['ip']
            except:
                pass

    def debug(self, msg):
        if self._show_debug:
            logger.warning(msg)

    def to_dict(self):
        data = {}
        for k, v in self.__dict__.items():
            if k.startswith('_'):
                continue
            if isinstance(v, datetime.datetime):
                data[k] = v.strftime("%Y-%m-%d %H:%M:%S")
            elif isinstance(v, datetime.timedelta):
                data[k] = v.total_seconds()
            elif k == 'data' and v:
                data[k] = json.dumps(v, ensure_ascii=False)
            else:
                data[k] = v
        return data

    def __repr__(self):
        return (
            f"SpiderInfo(pid={self.pid}, uid={self.uid}, name={self.name}, group={self.group_name}, "
            f"lan_ip={self.lan_ip}, wan_ip={self.wan_ip}, start_time={self.start_time}, "
            f"end_time={self.end_time}, run_time={self.run_time}, log_file={self.log_file}, "
            f"file={self.file}, interpreter={self.interpreter}, status={self.status}, "
            f"insert_count={self.insert_count}, progress={self.progress}, "
            f"total_progress={self.total_progress}, exception={self.exception})")

    def __monitor(self):
        num = 0
        while not self._stop_event.is_set():
            if datetime.date.today() > self._today:
                self.reset()
            result = self.notice()
            if not result:
                num += 1
                if num >= MAX_RETRY:
                    logger.error(f'上传数据失败，超过最大重试次数 {MAX_RETRY} 次，终止上报')
                    break
            else:
                num = 0
            time.sleep(MONITOR_INTERVAL)

    def notice(self):
        try:
            self.debug('上报数据')
            data = self.to_dict()
            self.debug(data)
            response = requests.post(url=self._monitor_endpoint, json=data, timeout=4)
            self.debug(response.text)
            return True
        except requests.exceptions.RequestException as e:
            self.debug(f'上报失败: {e}')
            return False

    def start_monitor(self):
        self.debug('启动监控')
        self._monitor_thread = threading.Thread(target=self.__monitor)
        self._monitor_thread.setDaemon(True)
        self._monitor_thread.start()

    def update_count(self, v: int = 1):
        with self._lock:
            self.insert_count += v

    def add_count(self, v: int = 1):
        with self._lock:
            self.insert_count += v
    def add_progress(self, v: int = 1):
        with self._lock:
            self.progress += v


    def stop(self):
        self._stop_event.set()
        self.__before_exit()

    def reset(self,msg=''):
        # 重置
        now = datetime.datetime.now()
        data = self.to_dict()
        data['run_time'] = (now - self.start_time).total_seconds()
        data['end_time'] = now.strftime("%Y-%m-%d %H:%M:%S")
        data['msg'] = data['msg'] + msg
        data['status'] = 2
        is_notice = False
        for _ in range(5):
            try:
                self.debug('上报数据reset')
                self.debug(data)
                response = requests.post(url=self._monitor_endpoint, json=data, timeout=4)
                self.debug(response.text)
                is_notice = True
                break
            except Exception as e:
                self.debug(f'上报失败: {e}')
                continue

        if is_notice:
            if not self.parent_uid:
                self.parent_uid = self.uid
            self.uid = str(uuid.uuid4())
            self.start_time = now
            self.insert_count = 0
            self._today = datetime.date.today()





import abc
import threading
import time
from concurrent.futures import ThreadPoolExecutor
from typing import Optional

from . import SpiderInfo
from .http import Request, BaseRequest


class SpiderThreadPool(object):

    def __init__(self):
        self.executor: Optional[ThreadPoolExecutor] = None

    def future_callback(self, future):
        if future.exception():
            raise future.exception()

    def submit_task(self, task_func, task):
        """
        向线程池中添加新任务
        :param task_func:
        :param task:
        :return:
        """
        future = self.executor.submit(task_func, task)
        future.add_done_callback(self.future_callback)

    def get_task_count(self) -> int:
        """
        获取当前线程池中还有多少任务数量
        :return:
        """
        return self.executor._work_queue.qsize()

    def start_batch_task(self, task_func, task_list: list, thread_num: int, wait=True):
        """
        多线程批量处理任务
        :param task_func: 任务函数
        :param task_list: 任务列表
        :param thread_num: 线程数量
        :param wait: 是否等待
        :return:
        """
        self.executor = ThreadPoolExecutor(max_workers=thread_num)
        try:
            for task in task_list:
                future = self.executor.submit(task_func, *task)
                future.add_done_callback(self.future_callback)
        finally:
            if wait:
                self.executor.shutdown(wait=True)
            # if wait:
            #     self.executor.shutdown(wait=False)
            #     while True:
            #         try:
            #             time.sleep(10)
            #         except KeyboardInterrupt:
            #             self.executor.shutdown(wait=True,cancel_futures=True)

    def task_wait(self):
        self.executor.shutdown(wait=True)


class BaseSpider(abc.ABC):

    def __init__(self, name, group_name=None, monitor=False, monitor_endpoint='http://127.0.0.1:14000/endpoint',
                 headers=None, cookies=None, proxy_url=None,task_type=0):
        self.info = SpiderInfo(name=name, group_name=group_name, monitor=monitor, monitor_endpoint=monitor_endpoint,task_type=task_type)
        self.local = threading.local()
        self.headers = headers if headers is not None else getattr(self.__class__, "headers", None)
        self.cookies = cookies if cookies is not None else getattr(self.__class__, "cookies", None)
        self.proxy_url = proxy_url if proxy_url is not None else getattr(self.__class__, "proxy_url", None)

    def get_request(self):
        return Request(proxy_url=self.proxy_url, headers=self.headers, cookies=self.cookies)

    @property
    def request(self) -> BaseRequest:
        if not hasattr(self.local, 'request'):
            self.local.request = self.get_request()
        return self.local.request

    def reset_monitor(self):
        self.info.stop()
        name = self.info.name
        group_name = self.info.group_name
        monitor_endpoint = self.info._monitor_endpoint
        del self.info
        self.info = SpiderInfo(name=name, group_name=group_name, monitor_endpoint=monitor_endpoint, monitor=True)

    def start(self, *args):
        pass

    def list(self, *args):
        pass

    def details(self, *args):
        pass

    def parse(self, *args):
        pass



from importlib import import_module
from typing import Any

class LazyLoader:
    def __init__(self, lib_name: str):
        self.lib_name = lib_name
        self._lib = None

    def __getattr__(self, name: str) -> Any:
        if self._lib is None:
            self._lib = import_module(self.lib_name)
        return getattr(self._lib, name)